@@ -45,10 +45,10 @@ struct ExpressionAnalyzer {
   // Given a stack of expressions, checks if the topmost is used as a result.
   // For example, if the parent is a block and the node is before the last position,
   // it is not used.
-  static bool isResultUsed(std::vector<Expression*> stack, Function* func);
+  static bool isResultUsed(ExpressionStack& stack, Function* func);
 
   // Checks if a value is dropped.
-  static bool isResultDropped(std::vector<Expression*> stack);
+  static bool isResultDropped(ExpressionStack& stack);
 
   // Checks if a break is a simple - no condition, no value, just a plain branching
   static bool isSimple(Break* curr) {
@@ -80,16 +80,26 @@ struct ExpressionAnalyzer {
   }
 
   // hash an expression, ignoring superficial details like specific internal names
-  static uint32_t hash(Expression* curr);
+  static HashType hash(Expression* curr);
 };
 
-// Re-Finalizes all node types
+// Re-Finalizes all node types. This can be run after code was modified in
+// various ways that require propagating types around, and it does such an
+// "incremental" update. This is done under the assumption that there is
+// a valid assignment of types to apply.
 // This removes "unnecessary' block/if/loop types, i.e., that are added
 // specifically, as in
 //  (block (result i32) (unreachable))
 // vs
 //  (block (unreachable))
 // This converts to the latter form.
+// This also removes un-taken branches that would be a problem for
+// refinalization: if a block has been marked unreachable, and has
+// branches to it with values of type unreachable, then we don't
+// know the type for the block: it can't be none since the breaks
+// exist, but the breaks don't declare the type, rather everything
+// depends on the block. To avoid looking at the parent or something
+// else, just remove such un-taken branches.
 struct ReFinalize : public WalkerPass<PostWalker<ReFinalize, OverriddenVisitor<ReFinalize>>> {
   bool isFunctionParallel() override { return true; }
 
@@ -102,108 +112,57 @@ struct ReFinalize : public WalkerPass<PostWalker<ReFinalize, OverriddenVisitor<R
 
   std::map<Name, Type> breakValues;
 
-  void visitBlock(Block* curr) {
-    if (curr->list.size() == 0) {
-      curr->type = none;
-      return;
-    }
-    // do this quickly, without any validation
-    auto old = curr->type;
-    // last element determines type
-    curr->type = curr->list.back()->type;
-    // if concrete, it doesn't matter if we have an unreachable child, and we
-    // don't need to look at breaks
-    if (isConcreteType(curr->type)) return;
-    // otherwise, we have no final fallthrough element to determine the type,
-    // could be determined by breaks
-    if (curr->name.is()) {
-      auto iter = breakValues.find(curr->name);
-      if (iter != breakValues.end()) {
-        // there is a break to here
-        auto type = iter->second;
-        if (type == unreachable) {
-          // all we have are breaks with values of type unreachable, and no
-          // concrete fallthrough either. we must have had an existing type, then
-          curr->type = old;
-          assert(isConcreteType(curr->type));
-        } else {
-          curr->type = type;
-        }
-        return;
-      }
-    }
-    if (curr->type == unreachable) return;
-    // type is none, but we might be unreachable
-    if (curr->type == none) {
-      for (auto* child : curr->list) {
-        if (child->type == unreachable) {
-          curr->type = unreachable;
-          break;
-        }
-      }
-    }
-  }
-  void visitIf(If* curr) { curr->finalize(); }
-  void visitLoop(Loop* curr) { curr->finalize(); }
-  void visitBreak(Break* curr) {
-    curr->finalize();
-    updateBreakValueType(curr->name, getValueType(curr->value));
-  }
-  void visitSwitch(Switch* curr) {
-    curr->finalize();
-    auto valueType = getValueType(curr->value);
-    for (auto target : curr->targets) {
-      updateBreakValueType(target, valueType);
-    }
-    updateBreakValueType(curr->default_, valueType);
-  }
-  void visitCall(Call* curr) { curr->finalize(); }
-  void visitCallIndirect(CallIndirect* curr) { curr->finalize(); }
-  void visitGetLocal(GetLocal* curr) { curr->finalize(); }
-  void visitSetLocal(SetLocal* curr) { curr->finalize(); }
-  void visitGetGlobal(GetGlobal* curr) { curr->finalize(); }
-  void visitSetGlobal(SetGlobal* curr) { curr->finalize(); }
-  void visitLoad(Load* curr) { curr->finalize(); }
-  void visitStore(Store* curr) { curr->finalize(); }
-  void visitAtomicRMW(AtomicRMW* curr) { curr->finalize(); }
-  void visitAtomicCmpxchg(AtomicCmpxchg* curr) { curr->finalize(); }
-  void visitAtomicWait(AtomicWait* curr) { curr->finalize(); }
-  void visitAtomicWake(AtomicWake* curr) { curr->finalize(); }
-  void visitConst(Const* curr) { curr->finalize(); }
-  void visitUnary(Unary* curr) { curr->finalize(); }
-  void visitBinary(Binary* curr) { curr->finalize(); }
-  void visitSelect(Select* curr) { curr->finalize(); }
-  void visitDrop(Drop* curr) { curr->finalize(); }
-  void visitReturn(Return* curr) { curr->finalize(); }
-  void visitHost(Host* curr) { curr->finalize(); }
-  void visitNop(Nop* curr) { curr->finalize(); }
-  void visitUnreachable(Unreachable* curr) { curr->finalize(); }
+  void visitBlock(Block* curr);
+  void visitIf(If* curr);
+  void visitLoop(Loop* curr);
+  void visitBreak(Break* curr);
+  void visitSwitch(Switch* curr);
+  void visitCall(Call* curr);
+  void visitCallIndirect(CallIndirect* curr);
+  void visitGetLocal(GetLocal* curr);
+  void visitSetLocal(SetLocal* curr);
+  void visitGetGlobal(GetGlobal* curr);
+  void visitSetGlobal(SetGlobal* curr);
+  void visitLoad(Load* curr);
+  void visitStore(Store* curr);
+  void visitAtomicRMW(AtomicRMW* curr);
+  void visitAtomicCmpxchg(AtomicCmpxchg* curr);
+  void visitAtomicWait(AtomicWait* curr);
+  void visitAtomicNotify(AtomicNotify* curr);
+  void visitSIMDExtract(SIMDExtract* curr);
+  void visitSIMDReplace(SIMDReplace* curr);
+  void visitSIMDShuffle(SIMDShuffle* curr);
+  void visitSIMDBitselect(SIMDBitselect* curr);
+  void visitSIMDShift(SIMDShift* curr);
+  void visitMemoryInit(MemoryInit* curr);
+  void visitDataDrop(DataDrop* curr);
+  void visitMemoryCopy(MemoryCopy* curr);
+  void visitMemoryFill(MemoryFill* curr);
+  void visitConst(Const* curr);
+  void visitUnary(Unary* curr);
+  void visitBinary(Binary* curr);
+  void visitSelect(Select* curr);
+  void visitDrop(Drop* curr);
+  void visitReturn(Return* curr);
+  void visitHost(Host* curr);
+  void visitNop(Nop* curr);
+  void visitUnreachable(Unreachable* curr);
 
-  void visitFunction(Function* curr) {
-    // we may have changed the body from unreachable to none, which might be bad
-    // if the function has a return value
-    if (curr->result != none && curr->body->type == none) {
-      Builder builder(*getModule());
-      curr->body = builder.blockify(curr->body, builder.makeUnreachable());
-    }
-  }
+  void visitFunction(Function* curr);
 
-  void visitFunctionType(FunctionType* curr) { WASM_UNREACHABLE(); }
-  void visitExport(Export* curr) { WASM_UNREACHABLE(); }
-  void visitGlobal(Global* curr) { WASM_UNREACHABLE(); }
-  void visitTable(Table* curr) { WASM_UNREACHABLE(); }
-  void visitMemory(Memory* curr) { WASM_UNREACHABLE(); }
-  void visitModule(Module* curr) { WASM_UNREACHABLE(); }
+  void visitFunctionType(FunctionType* curr);
+  void visitExport(Export* curr);
+  void visitGlobal(Global* curr);
+  void visitTable(Table* curr);
+  void visitMemory(Memory* curr);
+  void visitModule(Module* curr);
 
-  Type getValueType(Expression* value) {
-    return value ? value->type : none;
-  }
+private:
+  void updateBreakValueType(Name name, Type type);
 
-  void updateBreakValueType(Name name, Type type) {
-    if (type != unreachable || breakValues.count(name) == 0) {
-      breakValues[name] = type;
-    }
-  }
+  // Replace an untaken branch/switch with an unreachable value.
+  // A condition may also exist and may or may not be unreachable.
+  void replaceUntaken(Expression* value, Expression* condition);
 };
 
 // Re-finalize a single node. This is slow, if you want to refinalize
@@ -225,7 +184,16 @@ struct ReFinalizeNode : public OverriddenVisitor<ReFinalizeNode> {
   void visitAtomicRMW(AtomicRMW* curr) { curr->finalize(); }
   void visitAtomicCmpxchg(AtomicCmpxchg* curr) { curr->finalize(); }
   void visitAtomicWait(AtomicWait* curr) { curr->finalize(); }
-  void visitAtomicWake(AtomicWake* curr) { curr->finalize(); }
+  void visitAtomicNotify(AtomicNotify* curr) { curr->finalize(); }
+  void visitSIMDExtract(SIMDExtract* curr) { curr->finalize(); }
+  void visitSIMDReplace(SIMDReplace* curr) { curr->finalize(); }
+  void visitSIMDShuffle(SIMDShuffle* curr) { curr->finalize(); }
+  void visitSIMDBitselect(SIMDBitselect* curr) { curr->finalize(); }
+  void visitSIMDShift(SIMDShift* curr) { curr->finalize(); }
+  void visitMemoryInit(MemoryInit* curr) { curr->finalize(); }
+  void visitDataDrop(DataDrop* curr) { curr->finalize(); }
+  void visitMemoryCopy(MemoryCopy* curr) { curr->finalize(); }
+  void visitMemoryFill(MemoryFill* curr) { curr->finalize(); }
   void visitConst(Const* curr) { curr->finalize(); }
   void visitUnary(Unary* curr) { curr->finalize(); }
   void visitBinary(Binary* curr) { curr->finalize(); }
@@ -244,7 +212,7 @@ struct ReFinalizeNode : public OverriddenVisitor<ReFinalizeNode> {
   void visitModule(Module* curr) { WASM_UNREACHABLE(); }
 
   // given a stack of nested expressions, update them all from child to parent
-  static void updateStack(std::vector<Expression*>& expressionStack) {
+  static void updateStack(ExpressionStack& expressionStack) {
     for (int i = int(expressionStack.size()) - 1; i >= 0; i--) {
       auto* curr = expressionStack[i];
       ReFinalizeNode().visit(curr);
