@@ -75,6 +75,10 @@ struct CoalesceLocals : public WalkerPass<LivenessWalker<CoalesceLocals, Visitor
     interferences[low * numLocals + high] = 1;
   }
 
+  void unInterfere(Index i, Index j) {
+    interferences[std::min(i, j) * numLocals + std::max(i, j)] = 0;
+  }
+
   bool interferes(Index i, Index j) {
     return interferences[std::min(i, j) * numLocals + std::max(i, j)];
   }
@@ -271,7 +275,7 @@ std::vector<Index> adjustOrderByPriorities(std::vector<Index>& baseline, std::ve
     return priorities[x] > priorities[y] || (priorities[x] == priorities[y] && reversed[x] < reversed[y]);
   });
   return ret;
-};
+}
 
 void CoalesceLocals::pickIndices(std::vector<Index>& indices) {
   if (numLocals == 0) return;
@@ -279,11 +283,6 @@ void CoalesceLocals::pickIndices(std::vector<Index>& indices) {
     indices.push_back(0);
     return;
   }
-  if (getFunction()->getNumVars() <= 1) {
-    // nothing to think about here, since we can't reorder params
-    indices = makeIdentity(numLocals);
-    return;
-  }
   // take into account total copies. but we must keep params in place, so give them max priority
   auto adjustedTotalCopies = totalCopies;
   auto numParams = getFunction()->getNumParams();
@@ -315,25 +314,6 @@ void CoalesceLocals::pickIndices(std::vector<Index>& indices) {
   }
 }
 
-// Remove a copy from a set of an if, where one if arm is a get of the same set
-static void removeIfCopy(Expression** origin, SetLocal* set, If* iff, Expression*& copy, Expression*& other, Module* module) {
-  // replace the origin with the if, and sink the set into the other non-copying arm
-  bool tee = set->isTee();
-  *origin = iff;
-  set->value = other;
-  set->finalize();
-  other = set;
-  // if this is not a tee, then we can get rid of the copy in that arm
-  if (!tee) {
-    // we don't need the copy at all
-    copy = nullptr;
-    if (!iff->ifTrue) {
-      Builder(*module).flip(iff);
-    }
-    iff->finalize();
-  }
-}
-
 void CoalesceLocals::applyIndices(std::vector<Index>& indices, Expression* root) {
   assert(indices.size() == numLocals);
   for (auto& curr : basicBlocks) {
@@ -362,20 +342,6 @@ void CoalesceLocals::applyIndices(std::vector<Index>& indices, Expression* root)
           }
           continue;
         }
-        if (auto* iff = set->value->dynCast<If>()) {
-          if (auto* get = iff->ifTrue->dynCast<GetLocal>()) {
-            if (get->index == set->index) {
-              removeIfCopy(action.origin, set, iff, iff->ifTrue, iff->ifFalse, getModule());
-              continue;
-            }
-          }
-          if (auto* get = iff->ifFalse->dynCast<GetLocal>()) {
-            if (get->index == set->index) {
-              removeIfCopy(action.origin, set, iff, iff->ifFalse, iff->ifTrue, getModule());
-              continue;
-            }
-          }
-        }
       }
     }
   }
