@@ -42,7 +42,7 @@ using namespace wasm;
 struct DCENode {
   Name name;
   std::vector<Name> reaches; // the other nodes this one can reach
-  DCENode() {}
+  DCENode() = default;
   DCENode(Name name) : name(name) {}
 };
 
@@ -195,7 +195,9 @@ struct MetaDCEGraph {
       rooter.walk(segment.offset);
     }
     for (auto& segment : wasm.memory.segments) {
-      rooter.walk(segment.offset);
+      if (!segment.isPassive) {
+        rooter.walk(segment.offset);
+      }
     }
 
     // A parallel scanner for function bodies
