@@ -29,14 +29,25 @@ namespace BinaryConsts {
 namespace UserSections {
 const char* Name = "name";
 const char* SourceMapUrl = "sourceMappingURL";
-
 const char* Dylink = "dylink";
+const char* Linking = "linking";
+const char* Producers = "producers";
+const char* TargetFeatures = "target_features";
+const char* AtomicsFeature = "atomics";
+const char* BulkMemoryFeature = "bulk-memory";
+const char* ExceptionHandlingFeature = "exception-handling";
+const char* TruncSatFeature = "nontrapping-fptoint";
+const char* SignExtFeature = "sign-ext";
+const char* SIMD128Feature = "simd128";
 }
 }
 
 Name GROW_WASM_MEMORY("__growWasmMemory"),
-     MEMORY_BASE("memoryBase"),
-     TABLE_BASE("tableBase"),
+     WASM_CALL_CTORS("__wasm_call_ctors"),
+     MEMORY_BASE("__memory_base"),
+     TABLE_BASE("__table_base"),
+     GET_TEMP_RET0("getTempRet0"),
+     SET_TEMP_RET0("setTempRet0"),
      NEW_SIZE("newSize"),
      MODULE("module"),
      START("start"),
@@ -45,7 +56,7 @@ Name GROW_WASM_MEMORY("__growWasmMemory"),
      RESULT("result"),
      MEMORY("memory"),
      DATA("data"),
-     SEGMENT("segment"),
+     PASSIVE("passive"),
      EXPORT("export"),
      IMPORT("import"),
      TABLE("table"),
@@ -64,7 +75,7 @@ Name GROW_WASM_MEMORY("__growWasmMemory"),
      NEG_NAN("-nan"),
      CASE("case"),
      BR("br"),
-     ANYFUNC("anyfunc"),
+     FUNCREF("funcref"),
      FAKE_RETURN("fake_return_waka123"),
      MUT("mut"),
      SPECTEST("spectest"),
@@ -83,10 +94,10 @@ const char* getExpressionName(Expression* curr) {
     case Expression::Id::SwitchId: return "switch";
     case Expression::Id::CallId: return "call";
     case Expression::Id::CallIndirectId: return "call_indirect";
-    case Expression::Id::GetLocalId: return "get_local";
-    case Expression::Id::SetLocalId: return "set_local";
-    case Expression::Id::GetGlobalId: return "get_global";
-    case Expression::Id::SetGlobalId: return "set_global";
+    case Expression::Id::GetLocalId: return "local.get";
+    case Expression::Id::SetLocalId: return "local.set";
+    case Expression::Id::GetGlobalId: return "global.get";
+    case Expression::Id::SetGlobalId: return "global.set";
     case Expression::Id::LoadId: return "load";
     case Expression::Id::StoreId: return "store";
     case Expression::Id::ConstId: return "const";
@@ -101,9 +112,19 @@ const char* getExpressionName(Expression* curr) {
     case Expression::Id::AtomicCmpxchgId: return "atomic_cmpxchg";
     case Expression::Id::AtomicRMWId: return "atomic_rmw";
     case Expression::Id::AtomicWaitId: return "atomic_wait";
-    case Expression::Id::AtomicWakeId: return "atomic_wake";
-    default: WASM_UNREACHABLE();
+    case Expression::Id::AtomicNotifyId: return "atomic_notify";
+    case Expression::Id::SIMDExtractId: return "simd_extract";
+    case Expression::Id::SIMDReplaceId: return "simd_replace";
+    case Expression::Id::SIMDShuffleId: return "simd_shuffle";
+    case Expression::Id::SIMDBitselectId: return "simd_bitselect";
+    case Expression::Id::SIMDShiftId: return "simd_shift";
+    case Expression::Id::MemoryInitId: return "memory_init";
+    case Expression::Id::DataDropId: return "data_drop";
+    case Expression::Id::MemoryCopyId: return "memory_copy";
+    case Expression::Id::MemoryFillId: return "memory_fill";
+    case Expression::Id::NumExpressionIds: WASM_UNREACHABLE();
   }
+  WASM_UNREACHABLE();
 }
 
 // core AST type checking
@@ -406,9 +427,87 @@ void AtomicWait::finalize() {
   }
 }
 
-void AtomicWake::finalize() {
+void AtomicNotify::finalize() {
   type = i32;
-  if (ptr->type == unreachable || wakeCount->type == unreachable) {
+  if (ptr->type == unreachable || notifyCount->type == unreachable) {
+    type = unreachable;
+  }
+}
+
+void SIMDExtract::finalize() {
+  assert(vec);
+  switch (op) {
+    case ExtractLaneSVecI8x16:
+    case ExtractLaneUVecI8x16:
+    case ExtractLaneSVecI16x8:
+    case ExtractLaneUVecI16x8:
+    case ExtractLaneVecI32x4: type = i32; break;
+    case ExtractLaneVecI64x2: type = i64; break;
+    case ExtractLaneVecF32x4: type = f32; break;
+    case ExtractLaneVecF64x2: type = f64; break;
+    default: WASM_UNREACHABLE();
+  }
+  if (vec->type == unreachable) {
+    type = unreachable;
+  }
+}
+
+void SIMDReplace::finalize() {
+  assert(vec && value);
+  type = v128;
+  if (vec->type == unreachable || value->type == unreachable) {
+    type = unreachable;
+  }
+}
+
+void SIMDShuffle::finalize() {
+  assert(left && right);
+  type = v128;
+  if (left->type == unreachable || right->type == unreachable) {
+    type = unreachable;
+  }
+}
+
+void SIMDBitselect::finalize() {
+  assert(left && right && cond);
+  type = v128;
+  if (left->type == unreachable || right->type == unreachable || cond->type == unreachable) {
+    type = unreachable;
+  }
+}
+
+void MemoryInit::finalize() {
+  assert(dest && offset && size);
+  type = none;
+  if (dest->type == unreachable || offset->type == unreachable || size->type == unreachable) {
+    type = unreachable;
+  }
+}
+
+void DataDrop::finalize() {
+  type = none;
+}
+
+void MemoryCopy::finalize() {
+  assert(dest && source && size);
+  type = none;
+  if (dest->type == unreachable || source->type == unreachable || size->type == unreachable) {
+    type = unreachable;
+  }
+}
+
+void MemoryFill::finalize() {
+  assert(dest && value && size);
+  type = none;
+  if (dest->type == unreachable || value->type == unreachable || size->type == unreachable) {
+    type = unreachable;
+  }
+}
+
+void SIMDShift::finalize() {
+  assert(vec && shift);
+  type = v128;
+  if (vec->type == unreachable || shift->type == unreachable) {
     type = unreachable;
   }
 }
@@ -464,11 +563,19 @@ void Unary::finalize() {
     case TruncUFloat32ToInt32:
     case TruncSFloat64ToInt32:
     case TruncUFloat64ToInt32:
+    case TruncSatSFloat32ToInt32:
+    case TruncSatUFloat32ToInt32:
+    case TruncSatSFloat64ToInt32:
+    case TruncSatUFloat64ToInt32:
     case ReinterpretFloat32: type = i32; break;
     case TruncSFloat32ToInt64:
     case TruncUFloat32ToInt64:
     case TruncSFloat64ToInt64:
     case TruncUFloat64ToInt64:
+    case TruncSatSFloat32ToInt64:
+    case TruncSatUFloat32ToInt64:
+    case TruncSatSFloat64ToInt64:
+    case TruncSatUFloat64ToInt64:
     case ReinterpretFloat64: type = i64; break;
     case ReinterpretInt32:
     case ConvertSInt32ToFloat32:
@@ -480,7 +587,40 @@ void Unary::finalize() {
     case ConvertUInt32ToFloat64:
     case ConvertSInt64ToFloat64:
     case ConvertUInt64ToFloat64: type = f64; break;
-    default: std::cerr << "waka " << op << '\n'; WASM_UNREACHABLE();
+    case SplatVecI8x16:
+    case SplatVecI16x8:
+    case SplatVecI32x4:
+    case SplatVecI64x2:
+    case SplatVecF32x4:
+    case SplatVecF64x2:
+    case NotVec128:
+    case NegVecI8x16:
+    case NegVecI16x8:
+    case NegVecI32x4:
+    case NegVecI64x2:
+    case AbsVecF32x4:
+    case NegVecF32x4:
+    case SqrtVecF32x4:
+    case AbsVecF64x2:
+    case NegVecF64x2:
+    case SqrtVecF64x2:
+    case TruncSatSVecF32x4ToVecI32x4:
+    case TruncSatUVecF32x4ToVecI32x4:
+    case TruncSatSVecF64x2ToVecI64x2:
+    case TruncSatUVecF64x2ToVecI64x2:
+    case ConvertSVecI32x4ToVecF32x4:
+    case ConvertUVecI32x4ToVecF32x4:
+    case ConvertSVecI64x2ToVecF64x2:
+    case ConvertUVecI64x2ToVecF64x2: type = v128; break;
+    case AnyTrueVecI8x16:
+    case AllTrueVecI8x16:
+    case AnyTrueVecI16x8:
+    case AllTrueVecI16x8:
+    case AnyTrueVecI32x4:
+    case AllTrueVecI32x4:
+    case AnyTrueVecI64x2:
+    case AllTrueVecI64x2: type = i32; break;
+    case InvalidUnary: WASM_UNREACHABLE();
   }
 }
 
@@ -565,7 +705,6 @@ void Host::finalize() {
       }
       break;
     }
-    default: WASM_UNREACHABLE();
   }
 }
 
@@ -636,6 +775,17 @@ Type Function::getLocalType(Index index) {
   }
 }
 
+void Function::clearNames() {
+  localNames.clear();
+}
+
+void Function::clearDebugInfo() {
+  localIndices.clear();
+  debugLocations.clear();
+  prologLocation.clear();
+  epilogLocation.clear();
+}
+
 FunctionType* Module::getFunctionType(Name name) {
   auto iter = functionTypesMap.find(name);
   if (iter == functionTypesMap.end()) {
@@ -700,18 +850,20 @@ Global* Module::getGlobalOrNull(Name name) {
   return iter->second;
 }
 
-void Module::addFunctionType(FunctionType* curr) {
+FunctionType* Module::addFunctionType(std::unique_ptr<FunctionType> curr) {
   if (!curr->name.is()) {
     Fatal() << "Module::addFunctionType: empty name";
   }
   if (getFunctionTypeOrNull(curr->name)) {
     Fatal() << "Module::addFunctionType: " << curr->name << " already exists";
   }
-  functionTypes.push_back(std::unique_ptr<FunctionType>(curr));
-  functionTypesMap[curr->name] = curr;
+  auto* p = curr.get();
+  functionTypes.emplace_back(std::move(curr));
+  functionTypesMap[p->name] = p;
+  return p;
 }
 
-void Module::addExport(Export* curr) {
+Export* Module::addExport(Export* curr) {
   if (!curr->name.is()) {
     Fatal() << "Module::addExport: empty name";
   }
@@ -720,9 +872,11 @@ void Module::addExport(Export* curr) {
   }
   exports.push_back(std::unique_ptr<Export>(curr));
   exportsMap[curr->name] = curr;
+  return curr;
 }
 
-void Module::addFunction(Function* curr) {
+// TODO(@warchant): refactor all usages to use variant with unique_ptr
+Function* Module::addFunction(Function* curr) {
   if (!curr->name.is()) {
     Fatal() << "Module::addFunction: empty name";
   }
@@ -731,17 +885,33 @@ void Module::addFunction(Function* curr) {
   }
   functions.push_back(std::unique_ptr<Function>(curr));
   functionsMap[curr->name] = curr;
+  return curr;
 }
 
-void Module::addGlobal(Global* curr) {
+Function* Module::addFunction(std::unique_ptr<Function> curr) {
+  if (!curr->name.is()) {
+    Fatal() << "Module::addFunction: empty name";
+  }
+  if (getFunctionOrNull(curr->name)) {
+    Fatal() << "Module::addFunction: " << curr->name << " already exists";
+  }
+  auto* ret = functionsMap[curr->name] = curr.get();
+  functions.push_back(std::move(curr));
+  return ret;
+}
+
+Global* Module::addGlobal(Global* curr) {
   if (!curr->name.is()) {
     Fatal() << "Module::addGlobal: empty name";
   }
   if (getGlobalOrNull(curr->name)) {
     Fatal() << "Module::addGlobal: " << curr->name << " already exists";
   }
-  globals.push_back(std::unique_ptr<Global>(curr));
+
+  globals.emplace_back(curr);
+
   globalsMap[curr->name] = curr;
+  return curr;
 }
 
 void Module::addStart(const Name& s) {
@@ -809,4 +979,8 @@ void Module::updateMaps() {
   }
 }
 
+void Module::clearDebugInfo() {
+  debugInfoFileNames.clear();
+}
+
 } // namespace wasm
